/***************************************************************************
                          reko.cpp  -  description
                             -------------------
    begin                : Sun Aug 25 2002
    copyright            : (C) 2002 by kdereko@yahoo.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "reko.h"

#include <iostream.h>
#include <qfile.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qcolor.h>
#include <stdlib.h>
#include <math.h>
#include <qmap.h>
#include <qdir.h>
#include <qtextstream.h>

#define MAX_CARDS 200
#define KDE_X  72
#define KDE_Y  96

Reko::Reko(){



}
Reko::~Reko(){
}
/** get Bitnumber index from a byte array */
int Reko::getBit (char *data, unsigned long index){

  unsigned long byteno = index/8;
  char bitno = index & 7;
  bitno = 7-bitno;  // Bits are reversed

  char bytedata = data[byteno];
  char bitdata = bytedata >> bitno;

  return (bitdata & 1);
}
/** Swap card type order reko - kde
    As clubs,spades,hearts and diamonds are exchanged */
void Reko::assign_cards(int *map, int reko,int kde){
 map[reko-1]   = kde;
 map[reko+0] = kde+3;
 map[reko+1] = kde+2;
 map[reko+2] = kde+1;

}
/** Do the conversion from the REKO file infile to the
    KDE files outfile. If distort = 1 the aspect ratio
    is changed. Otherwise a border added */
bool Reko::convert(QString inname,QString outname,int distort)
{
  int map[MAX_CARDS];  // Map card numbers
/* order of cards:
   card 1		REKO production card
   card 2		empty card
   card 3		back of the cards
   cards 4..7	aces, all cards in order clubs, diamonds, hearts, spades
   cards 8..55	4 cards of 2,3,4,5,6,7,8,9,10,J,Q,K
   cards 56..59	4 stack background cards (left to right)
   cards 60..68	preferences cards
   		General, Prefs, Resolution, Select, Miscellaneous,
  		Music, Level, Confirm, About
*/
  for (int m=0;m<MAX_CARDS;m++)
  {
    map[m] = -1;  // Ignore card per default
  }
  map[3-1] = 0;    // Background

  // Exchange card numbers to suit KDE
  assign_cards(map,4,1);   // Aces
  assign_cards(map,8,49);  // 2
  assign_cards(map,12,45); // 3
  assign_cards(map,16,41); // 4
  assign_cards(map,20,37); // 5
  assign_cards(map,24,33); // 6
  assign_cards(map,28,29); // 7
  assign_cards(map,32,25); // 8
  assign_cards(map,36,21); // 9
  assign_cards(map,40,17); // 10
  assign_cards(map,44,13); // J
  assign_cards(map,48,9);  // Q
  assign_cards(map,52,5);  // K

  cout << "Start REKO conversion: " << inname << endl;

  // Open the given REKO file
  QFile file(inname);
  if (!file.open(IO_ReadOnly))
  {
    cout << "Cannot open " << inname << endl;
    return false;
  }

  struct RekoHeader header;
  struct Palette *palette;

  cout << "Size of header (22)=" <<sizeof(struct RekoHeader) << endl;
  cout << "Size of palette (3)=" <<sizeof(struct Palette) << endl;
  file.readBlock((char *)&header,sizeof(RekoHeader));

  if (memcmp(header.rh_ID,"REKO",4)!=0)
  {
    cout << "No REKO standard file" << endl;
    file.close();
    return false;
  }

  // Read header
  unsigned long rhMode = swapl(header.rh_ModeID);
  int rhCards  = header.rh_CardsCnt;
  int rhWidth  = swapw(header.rh_Width);
  int rhHeight = swapw(header.rh_Height);
  int rhDepth  = header.rh_Depth;
  int rhColors = (int)pow(2,rhDepth);
  unsigned long rhCardSize = swapl(header.rh_CardSize);
  unsigned long rhBodySize = swapl(header.rh_BodySize);
  int rhHam = (rhMode & 0x0800);

  cout << "Card number = " << rhCards << endl;
  cout << "Mode = " << hex << rhMode << " Ham=" << hex << rhHam << dec << endl;
  cout << "Bodysize=" << rhBodySize << " Cardsize=" << rhCardSize << endl;
  cout << "Width=" << rhWidth << " Height=" << rhHeight << endl;
  cout << "Depth=" << rhDepth << " Colors=" << rhColors << endl;

  // Always make 32 bit file
  //QImage image(rhWidth,rhHeight,32,rhColors);

  // Adjust to KDE aspect ratio
  int kdeWidth = (int)((float)rhHeight/4.0*3.0+0.5);
  int kdeOffset = (kdeWidth-rhWidth)/2;
  // If we sistort the image we keep the origial size
  // and change the aspect ratio later on
  if (distort)
  {
    kdeWidth = rhWidth;
    kdeOffset = 0;
  }
  QImage image(kdeWidth,rhHeight,32,rhColors);

  cout << "Aspect ratio corrected width=" << kdeWidth << " offset = " << kdeOffset << endl;

  cout << "image has " << image.numColors() << " colors" << endl;
  cout << "image depth " << image.depth() << endl;

  // Read palette
  palette=(struct Palette *)calloc(1,sizeof(struct Palette));
  if (!palette)
  {
    cout << "Cannot alloc palette" <<endl;
    file.close();
    return false;
  }

  // Fill with background color
  image.fill(image.color(0));

  int i;
  int palette_no = (rhHam?rhColors>>2:rhColors);
  cout << "Palette depth = " << palette_no <<endl;
  for (i=0;i<palette_no;i++)
  {
    file.readBlock((char *)palette,sizeof(struct Palette));
    QColor col=QColor(palette->r,palette->g,palette->b);
    image.setColor(i,col.rgb());
  }
  free(palette);

  char *data = (char *)calloc(1,rhCardSize);
  if (!data)
  {
    cout << "Cannot alloc card data" <<endl;
    file.close();
    return false;
  }

  // Read ll cards
  for (int card=0;card < rhCards; card ++)
  {
    file.readBlock((char *)data,rhCardSize);

    // Start the bitmap conversion
    unsigned int color;
    QRgb pixel_color;
    QRgb previous_color;
    int x,y,bitmap;
    // Loop y
    for (y=0;y<rhHeight;y++)
    {
      previous_color = image.color(0);  // background color for HAM
      // Loop x
      for (x=0;x<rhWidth;x++)
      {
        color = 0;
        // Loop bitmaps
        for (bitmap=0;bitmap<rhDepth;bitmap++)
        {
          unsigned long lineoffset = y*rhWidth*rhDepth;         // in bit
          unsigned long bitmapoffset = bitmap*rhWidth;          // in bit
          unsigned long bitno = lineoffset + bitmapoffset + x;  // in bit
          int bit = getBit(data,bitno);
          bit &= 1; // Make sure...
          color |= bit << bitmap;
        } // bitmap
        // Get special HAM color
        if (rhHam)
        {
          int ham_bits = color >> (rhDepth-2);            // high bits
          int pal_bits = color & ( (1<<(rhDepth-2)) -1);  // low bits
          int pal_shift = pal_bits << 2;
          int p_red   = qRed(previous_color);
          int p_green = qGreen(previous_color);
          int p_blue  = qBlue(previous_color);

          // HAM is a bit strange in its colors..assign them here
          switch(ham_bits)
          {
          case 0:
            pixel_color = image.color(pal_bits);
          break;
          case 1:
            pixel_color = qRgb(p_red,p_green,pal_shift);
          break;
          case 2:
            pixel_color = qRgb(pal_shift,p_green,p_blue);
          break;
          case 3:
            pixel_color = qRgb(p_red,pal_shift,p_blue);
          break;
          default:
            cout << "SERIOS error in HAM" << endl;
          break;
          }
        }  // end ham
        else
        {
          pixel_color = image.color(color);
        } // end normal

        // Finally set the data
        image.setPixel(x+kdeOffset,y,pixel_color);
        previous_color = pixel_color;
      } // x
    } // y

    // Adjust the picture to KDE sizes
    QImage scaled_image = image.smoothScale(KDE_X,KDE_Y,QImage::ScaleFree);

    if (card<MAX_CARDS)  // Ignore cards above our definition
    {
      // Remap the card numbers and write data
      int no = map[card];
      QString outfname;
      if (no == 0)  // background
      {
        outfname = outname+QString("back.png");
        scaled_image.save(outfname,"PNG");
      }
      else if (no>0)
      {
        outfname = outname+QString("%1.png").arg(no);
        scaled_image.save(outfname,"PNG");
      }
      // Otherwise just ignore

    }// end card ok
  }// end cards

  cout << "Image written to files" << endl;
  free(data);
  flush(cout);
  return true;
}
/** No descriptions */
bool Reko::makedir(QString name){
  QDir newdir;
  return newdir.mkdir(name);
}
/** No descriptions */
bool Reko::indexfile(QString filename,QString name){

  QFile file(filename);
  file.open(IO_WriteOnly);

  QTextStream t( &file );
  t << "[KDE Backdeck]" << endl;
  t << "Name = " << name << endl;
  t << "Preview = 11.png" << endl;
  t << "PySol =  false" << endl;
  t << "Comment =  Convert from REKO file " << name << endl;


  file.close();
}
